#!/usr/bin/python3

#----- imports -----

from .__init__ import __version__ as version, __doc__ as description
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from warnings import simplefilter
from sys import argv

from .__init__ import inform, warning, error
from .__init__ import longpath, shortpath, listfiles, listdirs, newbackfile, oldbackfile
from .__init__ import hold, chars, split, replace, shrink, findchar, rfindchar, chrs, ords
from .__init__ import retroenum, find, rfind
from os import remove, rename
from os.path import isfile

#----- global data -----

class Args: pass # container for arguments
class Glob: pass # container for other global data

#----- constants -----

CODE, TEXT, PICT = 'ctp' # values for out.records[jline][0] == line kind

#----- classes -----

class Buf:
    'paragraph buffer'

    def __init__(buf):
        buf.string = ''
        buf.indent = 0

    def __len__(buf):
        return len(buf.string)

    def append(buf, string, indent=0):
        if buf.string:
            buf.string += ' ' + shrink(string)
        else:
            buf.string = shrink(string)
            buf.indent = indent

    def flush(buf):
        if buf.string:
            prefix = (buf.indent - 2) * ' ' + '• ' if buf.indent else ''
            while len(buf.string) > Args.width - buf.indent:
                jchar = rfind(buf.string[:Args.width-buf.indent+1], ' ')
                if jchar <= 0:
                    error(f'Line {Glob.jline}, impossible to left-justify: {buf.string!r}')
                string, buf.string = buf.string[:jchar], buf.string[jchar+1:]
                if not Args.left_only: string = expand(string, Args.width - buf.indent)
                out.append('t', prefix + string)
                prefix = buf.indent * ' '
            if buf.string:
                out.append('t', prefix + buf.string)
                buf.string = ''

buf = Buf()

class Out:
    'output buffer'

    def __init__(out):
        out.records = []

    def __len__(out):
        return len(out.records)

    def append(out, kind, line):
        out.records.append((kind, line))

    def __call__(out, jline, jchar):
        try:
            if jline < 0 or jchar < 0: return ' '
            kind, line = out.records[jline]
            return line[jchar] if kind == PICT else ' '
        except IndexError:
            return ' '

    def pretty_lines(out):
        chstr = '#─│┐│┘│┤──┌┬└┴├┼'
        #        0123456789ABCDEF
        chset = frozenset(chstr)
        for jline, record in enumerate(out.records):
            kind, line = record
            if kind == PICT:
                chars = list(line)
                for jchar, char in enumerate(chars):
                    if char in chset:
                        chars[jchar] = chstr[1 * (out(jline, jchar - 1) in chset) +
                                             2 * (out(jline + 1, jchar) in chset) +
                                             4 * (out(jline - 1, jchar) in chset) +
                                             8 * (out(jline, jchar + 1) in chset)]
                out.records[jline] = (kind, ''.join(chars))

    def pretty_arrows(out):
        chstr = '@▷△@▽@@@◁@@@@@@@'
        #        0123456789ABCDEF
        chset = frozenset(chstr)
        for jline, record in enumerate(out.records):
            kind, line = record
            if kind == PICT:
                chars = list(line)
                for jchar, char in enumerate(chars):
                    if char in chset:
                        chars[jchar] = chstr[1 * (out(jline, jchar - 1) == '─') +
                                             2 * (out(jline + 1, jchar) == '│') +
                                             4 * (out(jline - 1, jchar) == '│') +
                                             8 * (out(jline, jchar + 1) == '─')]
                out.records[jline] = (kind, ''.join(chars))

    def pretty_chapters(out):
        jline_toc = -1
        jline_cp1 = -1
        out.chapters = []
        out.maxprefix = 0
        levels = []
        for jline, record in enumerate(out.records):
            kind, line = record
            if (kind == TEXT and line and line == line.upper() and
                (jline == 0 or not out.records[jline-1][1]) and
                (jline == len(out.records) - 1 or not out.records[jline+1][1])):
                if jline_toc < 0 and line[0] != line[0].lower(): # find content line
                    jline_toc = jline
                    out.records[jline] = (kind, shrink(line))
                else:
                    level = chapter_level(line) # find chapter lines
                    if level > 0:
                        if level > len(levels) + 1: error(f'Wrong chapter level sequence: {line!r}')
                        elif level == len(levels) + 1: levels.append(1)
                        else: levels = levels[:level]; levels[-1] += 1
                        prefix = '.'.join(str(n) for n in levels) + '. '
                        title = shrink(line.split(None, 1)[1])
                        out.maxprefix = max(out.maxprefix, len(prefix))
                        out.records[jline] = (TEXT, prefix + title)
                        out.chapters.append((prefix, title))
                        if jline_cp1 < 0: jline_cp1 = jline
        if jline_toc < 0: error('Table of content line not found')
        elif jline_cp1 < 0: error('Chapter title lines not found')
        elif jline_cp1 < jline_toc: error(f'Line {jline_cp1}: Chapter title line before content line: {out.records[jline_cp1][1]!r}')
        else:
            del out.records[jline_toc+1:jline_cp1] # delete old table of contents
            for prefix, title in [('','')] + out.chapters[::-1] + [('','')]: # insert new table of contents
                out.records.insert(jline_toc + 1, (PICT, Args.tab_blanks + prefix.ljust(out.maxprefix) + title.title()))

    def rewrite(out, file):
        with open(file, 'w') as file_w:
            for kind, line in out.records:
                print(line, file=file_w)
                if Args.print: print(line)                

out = Out()

#----- functions -----

def expand(string, width):
    'insert blanks into shrinked {string} until len({string}) == {width}'
    if len(string) >= width:
        return string
    if ' ' not in string[1:]:
        error(f'Line {Glob.jline}, impossible to right-justify: {buf.string!r}')
        return string
    chars = list(string)
    jchar = 0
    while True:
        if chars[jchar] != ' ' and chars[jchar + 1] == ' ':
            chars.insert(jchar + 1, ' ')
            if len(chars) == width:
                return ''.join(chars)
        jchar = jchar + 1 if jchar < len(chars) - 2 else 0

def chapter_level(line):
    'return 1 2 3... for chapter titles, 0 otherwise'
    status, level = 0, 0
    for char in line:
        if status == 0:
            if '0' <= char <= '9': status = 1
            else: break
        elif status == 1:
            if '0' <= char <= '9': status = 1
            elif char == '.': level += 1; status = 2
            else: break
        elif status == 2:
            if '0' <= char <= '9': status = 1
            elif char == ' ': return level
            else: break
    return 0

#----- main -----

def yawp(argv):
    'Yet Another Word Processor, an automatic word processor for text and Python files'

    parser = ArgumentParser(prog='yawp', formatter_class=RawDescriptionHelpFormatter, description=description)
    parser.add_argument('-V', '--version', action='version', version=f'yawp {version}')
    parser.add_argument('-v', '--verbose', action='store_true', help='show what happens')
    parser.add_argument('-u', '--undo', action='store_true', help='don\'t format, restore file to the previous version (default: backup and format)')
    parser.add_argument('-t', '--tab-blanks', type=int, default=4, help='blanks replacing a tab char in input (default: 4)')
    parser.add_argument('-w', '--width', type=int, default=89, help='output line width (default: 89)')
    parser.add_argument('-l', '--left-only', action='store_true', help='justify at left only (default: at left and right)')
    parser.add_argument('-g', '--pretty-graphics', action='store_true', help="replace '#' and '@' in pictures with lines and arrowheads")
    parser.add_argument('-c', '--pretty-chapters', action='store_true', help="renumber text chapters and refresh table of contents")
    parser.add_argument('-p', '--print', action='store_true', help='at end print formatted (or restored) file on stdout')
    parser.add_argument('file', help='file to be formatted (or restored if -u/--undo)')
    parser.parse_args(argv[1:], Args)
    Args.tab_blanks *= ' '
    Args.file = longpath(Args.file)
    
    if Args.undo: # restore: backfile --> Args.file
        backfile = oldbackfile(Args.file)
        if not backfile: error(f'Backup file for file {shortpath(Args.file)!r} not found')
        if Args.verbose: inform(f'Restore of file {shortpath(Args.file)!r} from file {shortpath(backfile)!r}')
        if isfile(Args.file): remove(Args.file)
        rename(backfile, Args.file)
        if Args.print:
            if Args.verbose: inform(f'Printing of file {shortpath(Args.file)!r}')
            for line in open(Args.file):
                print(line.replace('\t', Args.tab_blanks).rstrip())
    else:
        if not isfile(Args.file): error(f'File {shortpath(Args.file)!r} not found') # Args.file --> input
        input = [line.replace('\t', Args.tab_blanks).rstrip() for line in open(Args.file)]
        is_python_file = Args.file.endswith('.py')
        format = not is_python_file 
        
        for Glob.jline, line in enumerate(input): # input --> out.records
            is_switch_line = is_python_file and "\'\'\'" in line
            if is_switch_line: format = not format
            if is_switch_line or not format: # Python code
                buf.flush()
                out.append('c', line)
            elif not line: # empty line
                buf.flush()
                out.append('t', line)
            else:
                jdot = findchar(line, '[! ]')
                if jdot >= 0 and line[jdot:jdot+2] in ['• ','. ']: # dot line
                    buf.flush()
                    buf.append(line[jdot+2:], indent=jdot+2)
                elif line[0] == ' ': # indented line
                    if buf:
                        buf.append(line)
                    else:
                        if len(line) > Args.width: error(f'Line {Glob.jline} is too long: {line!r}')
                        out.append('p', line)
                else: # unindented line
                    buf.append(line)
        buf.flush()
        if is_python_file and format:
            error('Python file, odd number of switch lines')

        if Args.pretty_graphics:
            out.pretty_lines()
            out.pretty_arrows()
        if Args.pretty_chapters:
            out.pretty_chapters()

        backfile = newbackfile(Args.file) # backup: Args.file --> backfile
        if Args.verbose: inform(f'Backup of file {shortpath(Args.file)!r} into file {shortpath(backfile)!r}')
        rename(Args.file, backfile)

        if Args.verbose: inform(f'Rewriting file {shortpath(Args.file)!r}') # out.records --> Args.file
        if Args.verbose and Args.print: inform(f'Printing file {shortpath(Args.file)!r}')
        out.rewrite(Args.file)

def main():
    try:
        simplefilter('ignore')
        yawp(argv)
    except KeyboardInterrupt:
        print()

if __name__ == '__main__':
    main()

#----- end -----




#!/usr/bin/python3

'''Yet Another Word Processor, an automatic word processor for text and Python files

    I sound my barbaric yawp over the roofs of the world
                                          (Walt Whitman)

CONTENT

     1. Installation
     2. Help
     3. Introduction
     4. Logic
     5. Pretty Pictures
     6. Python Files
     7. An Example
     8. Messages
     9. History
    10. Author
    11. Arguments

1. INSTALLATION

Type from terminal:

    $ pip3 install yawp

2. HELP

Type from terminal:

    $ yawp -h

3. INTRODUCTION

The  name  "yawp"  here  means Yet Another Word Processor, and yawp is a simple automatic
word processor for text files and Python files, with the following features:

    • yawp makes a timestamped backup of file to be processed, allowing "undo" operation
    • yawp works "in place", reading formatting and rewriting the file to be processed
    • yawp operation is driven only by the content of the file and by a few parameters
    • yawp justifies "text" at left and right in:
        • unindented paragraphs
        • dot-marked indented paragraphs
    • yawp  accepts  unjustified "pictures" (as schemas, tables and code examples) freely
      intermixed with text
    • yawp's "pretty graphics" feature allows you to sketch draft pictures with lines and
      arrowheads  by  '#'  and '@' characters, which are automatically replaced by proper
      graphic characters
    • yawp's  "pretty  chapters"  feature ensures automatic multi-level chapter numbering
      and generation of table of contents
    • yawp  adopts  an ad hoc policy with Python files, formatting the docstrings but not
      the Python code
    • yawp  is  "stable",  if after a run of yawp on a file you run yawp a second time on
      the same file with the same arguments then the file content doesn't change

As  an  example,  this  documentation  you're  reading  has been formatted by yawp. Other
examples are scattered below.

4. LOGIC

"Pretty text" is an optional feature, it's activated by the -t/--pretty-text argument. If is not set,
then input lines are not justified, otherwise text justification takes place as follows.

Let's distinguish in four categories the file lines:

    • a line is an "empty line" if it contains no characters (note that in all input lines
      all  trailing  blanks  are  stripped  away,  hence every input line containing only
      blanks becomes an empty line)
    • otherwise a line is a "dot line" if the first nonblank character is a dot character
      '.' or '•' followed by a blank (on output such a '.' is always replaced by a '•')
    • otherwise a line is an "indented line" if it starts with a blank character
    • otherwise a line is an "unindented line"

More exactly, the "dot characters" are:

    • '.' is Unicode "decimal point", Python chr(46) or '\\x2e'
    • '•' is Unicode "black small circle", Python chr(8226) or '\\u2022'

The yawp algorithm, driven by the input lines, oscillates between two states:

    • "picture state", where input lines are directly written out as they are
    • "text state", where input lines are accumulated into a paragraph buffer for further
      justification and writing at paragraph end

The picture state is the initial state. In this state, if the input is:

    • an empty line or an indented line: the line is written out as is
    • an unindented line: text state is entered, an unindented paragraph begins, the line
      is  shrinked  (initial  and  multiple  blanks  are  eliminated) and assigned to the
      paragraph buffer, paragraph left indentation is set to zero
    • a  dot  line:  text  state  is  entered,  an indented paragraph begins, the line is
      shrinked and assigned to the paragraph buffer, paragraph left indentation is set to
      the position of initial dot character plus two
    • end of input file: processing is terminated

When we are in text state, if the input line is:

    • an  empty  line:  the  paragraph  buffer  is  flushed  (justified,  written out and
      emptied), state goes back to picture state, the empty line is written out
    • an  indented or unindented line: the line is shrinked and appended to the paragraph
      buffer
    • a  dot  line:  paragraph buffer is flushed, a new paragraph is started, the line is
      shrinked and assigned to the paragraph buffer, paragraph left indentation is set to
      the position of initial dot plus two
    • end of input file: paragraph buffer is flushed, processing is ended

For sake of clarity, the following diagram illustrates states and transitions.

        empty line,                                                 unindented line,
        indented line:                                              indented line:
        write                                                       append
        ┌──────────┐                                                ┌──────────┐
        │          │                                                │          │
        │          │                                                │          │
        │    ┌─────┴─────┐ unindented line, dot line: assign  ┌─────┴─────┐    │
        └───▷│           ├───────────────────────────────────▷│           │◁───┘
             │  picture  │      empty line: flush, write      │  text     │
   ─────────▷│           │◁───────────────────────────────────┤           │
             │  state    │                                    │  state    │
             │           │                             ┌──────┤           │◁───┐
             └─────┬─────┘                             │      └─────┬─────┘    │
                   │                                   │            │          │
                   │                                   │            │          │
        ◁──────────┘                        ◁──────────┘            └──────────┘
        end of file:                        end of file:            dot line:
        stop                                flush, stop             flush, assign

Actions associated to transitions are:

    • "write": the input line is immediately written out as is
    • "assign":  the  input  line  is  shrinked  (initial,  multiple and final blanks are
      removed) and assigned to the paragraph buffer
    • "append": the input line is shrinked and appended to the paragraph buffer
    • "flush": the paragraph buffer is flushed (justified, written out and emptied)
    • "stop": execution is finished

5. PRETTY GRAPHICS

"Pretty graphics" is an optional feature, it's activated by the -g/--pretty-graphics argument.

In  a  picture  you  can  sketch  draft  pictures  with  lines  and arrows by two special
characters:

    • use '#' to draw horizontal and vertical lines
    • use '@' to mark an arrowhead

Each  such  character  is  (possibly)  replaced  by yawp with a proper graphic character,
depending  on  the  other  four  characters  around (over, under, at left, and at right).
Isolated characters are not replaced.

This feature is active in image mode only, so it does not work in paragraphs.

An example:

    $ cat graphics.txt
           #############      
           # x - y     #      
         x #   #############               ######
           #   # x & y #   #               #N#N2#
           #############   # y             ######
               #     y - x #               #0# 0#
               #############               #1# 1#
                                           #2# 4#
                #####     #####            #3# 9#
          #####@# A #####@# B #@#####      #4#16#
                #####     #####     #      #5#25#
                  @         #       #      #6#36#
                  #         #       #      #7#49#
                  #         @       #      #8#64#
                #####     #####     #      #9#81#
          @###### D #@##### C #@#####      ######
                #####     #####
    $ yawp -v -w55 -g -p graphics.txt
    ... Backup of file '~/graphics.txt' into file '~/graphics-2022.01.03-17.14.27.txt'
    ... Rewriting file '~/graphics.txt'
    ... Printing file '~/graphics.txt'
           ┌───────────┐
           │ x - y     │
         x │   ┌───────┼───┐               ┌─┬──┐
           │   │ x & y │   │               │N│N2│
           └───┼───────┘   │ y             ├─┼──┤
               │     y - x │               │0│ 0│
               └───────────┘               │1│ 1│
                                           │2│ 4│
                ┌───┐     ┌───┐            │3│ 9│
          ─────▷│ A ├────▷│ B │◁────┐      │4│16│
                └───┘     └─┬─┘     │      │5│25│
                  △         │       │      │6│36│
                  │         │       │      │7│49│
                  │         ▽       │      │8│64│
                ┌─┴─┐     ┌───┐     │      │9│81│
          ◁─────┤ D │◁────┤ C │◁────┘      └─┴──┘
                └───┘     └───┘
    $ 

5. PRETTY CHAPTERS

"Pretty chapters" is an optional feature, it's activated by the -c/--pretty-chapters argument.
If set, it ensures automatic multi-level chapter numbering
      and generation of table of contents.

If pretty chapters is active, the file must contain:

    . a "content line", containing the title of the table of content
    . one or more "chapter lines", containing a multi-level numbering and a chapter title

A line is a content line if:

    . is the first line or is preceded by an empty line
    . is followed by an empty line
    . starts with an uppercase letter
    . contains uppercase letters and blanks only

A content line must precede all chapter lines. Examples:

    . 'CONTENT'
    . 'TABLE OF CONTENT'
    . 'INDEX OF CHAPTERS'

A line is a chapter line if:

    . is preceded by an empty line
    . is the last line or is followed by an empty line
    . contains:
        . one or more unsigned decimal integer constants, each followed by a '.' dot
        . a blank
        . a chapter title, containig any character but underscore letters

The "level" of chapter line is the count of number-dot couples in its prefix, examples:

    . '12345. A LEVEL-1 CHAPTER LINE'
    . '1.345. A LEVEL-2 CHAPTER LINE'
    . '0.0.0. A LEVEL-3 CHAPTER LINE'

Chapter lines must follow two sequence rules:

    . first chapter line must be a level-1 chapter line
    . each other chapter line can have a level between 1 and the level of the previous chapter line plus 1

Numbers in input don't matter, they are replaced by the right ones, only the level matters.

Lines between the content line and the first chapter line are suppose to contain the old table of content, hence they are deleted and replaced by the new automatically
generated table of content.

This feature is active in text mode only, so it does not work in pictures. For what has been said we can observe that each content line or chapter line must be a one-line unindented paragraph.

An example:

    $ cat chapters.txt
        TITLE OF DOCUMENT

    TABLE OF CONTENTS

        (old table of content
         will be replaced
         by the new one)

    0. AAA AAA

    ...

    32.33. BBB BBB

    0.0. CCC CCC

    0. DDD DDD

    0.0. EEE EEE

    0.0.0. FFF FFF

    0.0. GGG GGG
    $ yawp -v -w55 -c -p chapters.txt
    ... Backup of file '~/chapters.txt' into file '~/chapters-2022.01.03-17.01.14.txt'
    ... Rewriting file '~/chapters.txt'
    ... Printing file '~/chapters.txt'
        TITLE OF DOCUMENT

    TABLE OF CONTENTS

        1. Aaa Aaa
        1.1. Bbb Bbb
        1.2. Ccc Ccc
        2. Ddd Ddd
        2.1. Eee Eee
        2.1.1. Fff Fff
        2.2. Ggg Ggg

    1. AAA AAA

    ...

    1.1. BBB BBB

    1.2. CCC CCC

    2. DDD DDD

    2.1. EEE EEE

    2.1.1. FFF FFF

    2.2. GGG GGG
    $ 

6. PYTHON FILES

Python  files  deserve  a  special  treatment.  If  the textfile filename ends with '.py'
extension, then we suppose the file is a Python source, hence we are interested to format
docstrings and not Python code. So the formatting function is alternatively turned on and
off by switch lines. A "switch line" is a line containing a "\'\'\'" string.

Note that yawp never formats switch lines, formatting takes place from the line after the
"on" switch line until the line before the next "off" switch line.

So your Python file must follow some simple rules:

    • docstrings to be formatted must start and ended by "\'\'\'" and not '"""'
    • long strings not to be formatted must start and end with '"""' and not "\'\'\'"
    • a "\'\'\'" inside a string must be coded as "\\'\\'\\'"

An  error  in switch lines could format and destroy your Python code. A preliminary check
prints an error message and stops execution before file formatting if the total number of
switch  lines  is  odd. This should intercept 90% of errors, anyway after yawp processing
check the result and if needed go back to previous version by the -u/--undo yawp option.

An example:

    $ cat pycode.py
    #!/usr/bin/python3

    \'\'\' Text in "on" switch line is not formatted.
    This is a one-line unindented paragraph.

    This is a multi-line unindented paragraph.
    This is a multi-line unindented paragraph.
    This is a multi-line unindented paragraph.

        This is a picture, it remains as is.
                This is a picture, it remains as is.
            This is a picture, it remains as is.

        . This is a multi-line indented paragraph.
    This is a multi-line indented paragraph.
        This is a multi-line indented paragraph.
            . This is another multi-line indented paragraph.
    This is another multi-line indented paragraph.
        This is another multi-line indented paragraph.
    \'\'\' # Text in "off" switch line is not formatted.

    def double(x): # Python code is not formatted.
        \'\'\'
    This is another multi-line unindented paragraph.
    This is another multi-line unindented paragraph.
    This is another multi-line unindented paragraph.
    \'\'\'
        return x + x # Python code is not formatted.
    $ yawp -v -w55 -p pycode.py
    ... Backup of file '~/pycode.py' into file '~/pycode-2022.01.03-16.49.27.py'
    ... Rewriting file '~/pycode.py'
    ... Printing file '~/pycode.py'
    #!/usr/bin/python3

    \'\'\' Text in "on" switch line is not formatted.
    This is a one-line unindented paragraph.

    This  is  a  multi-line unindented paragraph. This is a
    multi-line  unindented  paragraph. This is a multi-line
    unindented paragraph.

        This is a picture, it remains as is.
                This is a picture, it remains as is.
            This is a picture, it remains as is.

        • This  is a multi-line indented paragraph. This is
          a   multi-line  indented  paragraph.  This  is  a
          multi-line indented paragraph.
            • This    is    another   multi-line   indented
              paragraph.   This   is   another   multi-line
              indented    paragraph.    This   is   another
              multi-line indented paragraph.
    \'\'\' # Text in "off" switch line is not formatted.

    def double(x): # Python code is not formatted.
        \'\'\'
    This  is  another multi-line unindented paragraph. This
    is  another  multi-line  unindented  paragraph. This is
    another multi-line unindented paragraph.
    \'\'\'
        return x + x # Python code is not formatted.
    $ 

8. MESSAGES

All  messages  are  written  on  stderr,  in  order to avoid interference with -p/--print
option, which writes on stdout.

There are two types of messages:
    • "information messages" say what's going on if -v/--verbose argument is set:
        • ... Backup of file '...' into file '...'
        • ... Restore of file '...' from file '...'
        • ... Processing of file '...'
        • ... Printing of file '...'
    • "error  messages"  are  written  when  execution  must be interrupted, in this case
      backup is not performed and file is not rewritten:
        • !!! File '...' not found, program halted
        • !!! Backup file for file '...' not found', program halted
        • !!! Line ..., impossible to left-justify: '...', program halted
        • !!! Line ..., impossible to right-justify: '...', program halted
        • !!! Line ... is too long: '...', program halted
        • !!! Python file, odd number of switch lines, program halted

9. HISTORY

    • version 0.4.1
        • first version published on pypi.org

10. AUTHOR

Written by Carlo Alessandro Verre, carlo.alessandro.verre@gmail.com.

11. ARGUMENTS

If -u is set then -w -l -g and -c are allowed but have no effect.

Note that -g and -c are indipendent, each can be set or not.
'''

__version__ = '0.4.1'

__all__ = []

#----- imports -----

from fnmatch import fnmatchcase
from glob import glob
from os.path import join as joinpath, split as splitpath, abspath, expanduser, splitext, exists, isfile, isdir
from sys import exit, stderr
from time import localtime

#----- message functions -----

def inform(message):
    'print an information message and continue'
    print(f'... {message}', file=stderr)

def warning(message):
    'print a warning message and continue'
    print(f'??? {message}', file=stderr)

def error(message):
    'print an error message and exit'
    exit(f'!!! {message}, program halted')

#----- path & file functions -----

def longpath(path='.'):
    return abspath(expanduser(path))

def shortpath(path='.'):
    path, home = longpath(path), longpath('~')
    return '~' + path[len(home):] if path.startswith(home) else path

def listfiles(pattern):
    "return a list of absolute paths to {pattern}-matching files, '**' in path is allowed"
    return [file for file in glob(longpath(pattern), recursive=True) if isfile(file)]

def listdirs(pattern):
    "return a list of absolute paths to {pattern}-matching dirs, '**' in path is allowed"
    return [dir for dir in glob(longpath(pattern), recursive=True) if isdir(dir)]

def newbackfile(file):
    'create a timestamped filename for backup of {file}'
    path, ext = splitext(longpath(file))
    return f'{path}-%04d.%02d.%02d-%02d.%02d.%02d{ext}' % localtime()[:6]

def oldbackfile(file):
    "return filename of the more recent timestamped backup of file {file}, or '' if not found"
    path, ext = splitext(longpath(file))
    return max(listfiles(f'{path}-????.??.??-??.??.??{ext}') or [''])

#----- string functions -----

def hold(string, charpattern, default=''):
    'hold {charpattern}-matching chars in {string} and replace not matching chars by {default}'
    return ''.join(char if fnmatchcase(char, charpattern) else default for char in string)

def chars(charpattern):
    'return a sorted string of all {charpattern}-matching characters'
    kernel = charpattern[1:-1]
    return ''.join(chr(j) for j in range(ord(min(kernel)), ord(max(kernel)) + 1)
                   if fnmatchcase(chr(j), charpattern))

def split(string, separator=None):
    'return [] if not string else string.split(separator)'
    return [] if not string else string.split(separator)

def replace(string, *oldsnews):
    'replace(string, a, b, c, d, ...) == string.replace(a, b).replace(c, d)...'
    for j in range(0, len(oldsnews), 2):
        string = string.replace(oldsnews[j], oldsnews[j+1])
    return string

def shrink(string, joinby=' ', splitby=None):
    'eliminate from {string} leading, multiple and trailing blanks'
    return joinby.join(string.split(splitby))

def findchar(string, pattern):
    'return index of first {pattern}-matching char found in {string}, -1 if not found'
    for j, char in enumerate(string):
        if fnmatchcase(char, pattern):
            return j
    else:
        return -1

def rfindchar(string, pattern):
    'return index of last {pattern}-matching char found in {string}, -1 if not found'
    for j, char in retroenum(string):
        if fnmatchcase(char, pattern):
            return j
    else:
        return -1

def chrs(jj):
    "return ''.join(chr(j) for j in jj)"
    return ''.join(chr(j) for j in jj)

def ords(string):
    'return [ord(char) for char in string]'
    return [ord(char) for char in string]

#----- sequence functions -----

def retroenum(sequence):
    'like enumerate(sequence), but backwards from last to first item'
    for j in range(len(sequence) - 1, -1, -1):
        yield j, sequence[j]

def find(sequence, item):
    'return index of first {item} found in {sequence}, -1 if not found'
    for j, itemj in enumerate(sequence):
        if itemj == item:
            return j
    else:
        return -1

def rfind(sequence, item):
    'return index of last {item} found in {sequence}, -1 if not found'
    for j, itemj in retroenum(sequence):
        if itemj == item:
            return j
    else:
        return -1

#----- end -----
